<!--Widget with a large number with text below. Useful for big bold stats-->
<div class="col-sm-2 col-xs-6 text-center">
  <div class="hero-widget well well-sm">
    <div class="icon">
      <i class="glyphicon glyphicon-{{icon}}"></i>
    </div>
    <div class="text">
      <var>{{number}}</var>
      <label class="text-muted">{{title}}</label>
    </div>
    <!-- 
    <div class="options">
      <a href="javascript:;" class="btn btn-primary btn-lg"><i class="glyphicon glyphicon-plus"></i> Add a guest</a>
    </div>
    --> 
  </div>
</div>
